__author__ = "IncognitoCoding"
__copyright__ = "Copyright 2022, exceptions"
__credits__ = ["IncognitoCoding"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "IncognitoCoding"
__status__ = "Beta"


class InputFailure(Exception):
    """Exception raised for an input exception message."""

    __module__ = "builtins"
    pass


class InvalidKeyError(Exception):
    """Exception raised for an invalid dictionary key."""

    __module__ = "builtins"
    pass
